import time
from datetime import datetime

from pandas import DataFrame, concat

from offline_results.common.config import (
    CONFIG_HOMEPAGE_PAYTV,
    CONFIG_HOMEPAGE_NO_PAYTV, LIVE_TV_CHANNEL_HOMEPAGE_PAYTV, LIVE_TV_CHANNEL_HOMEPAGE_NO_PAYTV,
)
from offline_results.common.constants import (
    PAY_TV_CONTENT,
    NO_PAY_TV_CONTENT,
    CONTENT_ID,
    HOMEPAGE_ID,
    MODEL_POPULAR,
    SERVICE_NAME,
    ALL_CONTENT_BASED,
    DEFAULT_POPULAR_MODULE_NAME,
    RECORDS,
    PAY_TV,
    LEFT,
    ACTIVE,
    CONTENT_STATUS,
    IS_PAY_TV,
    CREATED_ON,
    VIEW_COUNT,
    SUM,
    HOMEPAGE_ID_BASED,
    REC_TYPE,
    NO_PAY_TV,
    POPULAR_MODULE_NAME,
)
from offline_results.recommendation.homepage_recommendation.trending.trending_model import (
    TrendingModel,
)

from offline_results.recommendation.utils import RecommendationUtils
from offline_results.repository.graph_db_connection import ANGraphDb
from offline_results.utils import custom_exception, Logging


class PopularModelFallback:
    viewed_relation_history_df = DataFrame()
    live_tv_channel_paytv = []
    live_tv_channel_nopaytv = []

    @staticmethod
    @custom_exception()
    def get_data():
        obj = PopularModelFallback
        if len(obj.viewed_relation_history_df) != 0:
            return
        obj.viewed_relation_history_df = RecommendationUtils.user_viewed_data_from_s3()
        obj.live_tv_channel_paytv = RecommendationUtils.get_tv_channels(is_pay_tv=True)
        obj.live_tv_channel_nopaytv = RecommendationUtils.get_tv_channels(is_pay_tv=False)

    @staticmethod
    @custom_exception()
    def clear_data():
        obj = PopularModelFallback
        obj.viewed_relation_history_df = DataFrame()
        obj.live_tv_channel_paytv = []
        obj.live_tv_channel_nopaytv = []

    @staticmethod
    @custom_exception()
    def get_homepage_for_contents(is_pay_tv) -> DataFrame:
        try:
            query = f"""g.V().hasLabel('{PAY_TV_CONTENT if is_pay_tv else NO_PAY_TV_CONTENT}' ).match(
                __.as("c").values().as("content_id"),
                __.as("c").out("HAS_HOMEPAGE")
                .values("homepage_id").as("homepage_id")
                ).select("content_id", "homepage_id")"""
            graph = ANGraphDb.new_connection_config().graph
            data = graph.custom_query(
                query,
                payload={
                    PAY_TV_CONTENT: PAY_TV_CONTENT,
                    NO_PAY_TV_CONTENT: NO_PAY_TV_CONTENT,
                },
            )
            data = [rec for idx in data for rec in idx]
            df = DataFrame(data)
            return df
        except Exception:
            graph = ANGraphDb.new_connection_config().graph
            query = f"""g.V().hasLabel('{PAY_TV_CONTENT if is_pay_tv else NO_PAY_TV_CONTENT}' ).match(
                            __.as("c").values().as("content_id"),
                            __.as("c").out("HAS_HOMEPAGE")
                            .values("homepage_id").as("homepage_id")
                            ).select("content_id", "homepage_id")"""

            data = graph.custom_query(
                query,
                payload={
                    PAY_TV_CONTENT: PAY_TV_CONTENT,
                    NO_PAY_TV_CONTENT: NO_PAY_TV_CONTENT,
                },
            )
            data = [rec for idx in data for rec in idx]
            df = DataFrame(data)
            return df

    @staticmethod
    @custom_exception()
    def default_popular_all_content_algorithm(pay_tv_label):
        df = PopularModelFallback.viewed_relation_history_df
        label = True if pay_tv_label == PAY_TV else False
        df_pay_tv = df[(df[IS_PAY_TV] == label) & (df[CONTENT_STATUS] == ACTIVE)]
        try:
            graph = ANGraphDb.new_connection_config().graph
            Logging.info("Fetch content-homepage_id mapping")
            (
                content_homepage_id_mapping,
                content_homepage_df,
            ) = TrendingModel.get_content_homepage_id_mapping(
                graph, pay_tv_label, homepage_id_wise=False
            )
        except Exception:
            graph = ANGraphDb.new_connection_config().graph
            Logging.info("Re-trying to Fetch content-homepage_id mapping")
            (
                content_homepage_id_mapping,
                content_homepage_df,
            ) = TrendingModel.get_content_homepage_id_mapping(
                graph, pay_tv_label, homepage_id_wise=False
            )
        df_pay_tv = df_pay_tv.merge(
            content_homepage_id_mapping, on=CONTENT_ID, how=LEFT
        )
        df_pay_tv = df_pay_tv[~df_pay_tv[HOMEPAGE_ID].isna()].reset_index(drop=True)
        df_pay_tv = df_pay_tv.explode(column=HOMEPAGE_ID)
        df_popular = DataFrame(
            CONFIG_HOMEPAGE_PAYTV.items()
            if pay_tv_label == PAY_TV
            else CONFIG_HOMEPAGE_NO_PAYTV.items(),
            columns=[HOMEPAGE_ID, MODEL_POPULAR],
        )
        df_popular = df_popular[df_popular[MODEL_POPULAR] == POPULAR_MODULE_NAME.lower()]
        list_dedicated_homepage_id = df_popular[HOMEPAGE_ID].to_list()
        live_tv_channel_homepage = (
            LIVE_TV_CHANNEL_HOMEPAGE_PAYTV if pay_tv_label == PAY_TV else LIVE_TV_CHANNEL_HOMEPAGE_NO_PAYTV
        )
        list_dedicated_homepage_id.extend(live_tv_channel_homepage)
        df_pay_tv = df_pay_tv[df_pay_tv[HOMEPAGE_ID].isin(list_dedicated_homepage_id)]
        df_pay_tv = df_pay_tv[
            ~df_pay_tv[CONTENT_ID].isin(PopularModelFallback.live_tv_channel_paytv if pay_tv_label == PAY_TV else
                                        PopularModelFallback.live_tv_channel_nopaytv)
        ]
        df_sum = df_pay_tv.groupby(CONTENT_ID, as_index=False)[VIEW_COUNT].agg(SUM)
        df_sum = df_sum.sort_values(by=VIEW_COUNT, ascending=False).iloc[:20, :]
        df_sum = df_sum.merge(
            PopularModelFallback.get_homepage_for_contents(label),
            on=CONTENT_ID,
            how=LEFT,
        )
        df_sum.drop(columns=VIEW_COUNT, axis=1, inplace=True)
        df_sum.drop_duplicates(CONTENT_ID, inplace=True)
        df_sum[CREATED_ON] = datetime.utcnow().isoformat()
        df_sum[REC_TYPE] = DEFAULT_POPULAR_MODULE_NAME.upper()
        all_content_based_rec = {}
        key_prefix = (
            SERVICE_NAME
            + ":"
            + DEFAULT_POPULAR_MODULE_NAME.lower()
            + ":"
            + pay_tv_label
            + ":"
            + ALL_CONTENT_BASED
        )
        result = df_sum.to_dict(orient=RECORDS)
        all_content_based_rec[key_prefix] = result
        return all_content_based_rec

    @staticmethod
    @custom_exception()
    def default_popular_homepage_id_algorithm(user_label):
        df = PopularModelFallback.viewed_relation_history_df
        label = True if user_label == PAY_TV else False
        df_pay_tv = df[(df[IS_PAY_TV] == label) & (df[CONTENT_STATUS] == ACTIVE)]
        unique_ubd_content_id = df_pay_tv[CONTENT_ID].unique().tolist()
        try:
            graph = ANGraphDb.new_connection_config().graph
            Logging.info("Fetch content-homepage_id mapping")
            (
                content_homepage_id_mapping,
                content_homepage_df,
            ) = TrendingModel.get_content_homepage_id_mapping(
                graph, user_label, homepage_id_wise=True
            )
        except Exception:
            graph = ANGraphDb.new_connection_config().graph
            Logging.info("Re-trying to Fetch content-homepage_id mapping")
            (
                content_homepage_id_mapping,
                content_homepage_df,
            ) = TrendingModel.get_content_homepage_id_mapping(
                graph, user_label, homepage_id_wise=True
            )
        df_pay_tv = df_pay_tv.merge(
            content_homepage_id_mapping, on=CONTENT_ID, how=LEFT
        )
        df_pay_tv = df_pay_tv[~df_pay_tv[HOMEPAGE_ID].isna()].reset_index(drop=True)
        df_pay_tv = df_pay_tv.explode(column=HOMEPAGE_ID)
        df_popular = DataFrame(
            CONFIG_HOMEPAGE_PAYTV.items()
            if user_label == PAY_TV
            else CONFIG_HOMEPAGE_NO_PAYTV.items(),
            columns=[HOMEPAGE_ID, MODEL_POPULAR],
        )
        df_popular = df_popular[df_popular[MODEL_POPULAR] == POPULAR_MODULE_NAME.lower()]
        list_dedicated_homepage_id = df_popular[HOMEPAGE_ID].to_list()
        live_tv_channel_homepage = (
            LIVE_TV_CHANNEL_HOMEPAGE_PAYTV if user_label == PAY_TV else LIVE_TV_CHANNEL_HOMEPAGE_NO_PAYTV
        )
        list_dedicated_homepage_id.extend(live_tv_channel_homepage)
        df_pay_tv = df_pay_tv[df_pay_tv[HOMEPAGE_ID].isin(list_dedicated_homepage_id)]
        homepage_based_rec = {}
        for homepage_id in list_dedicated_homepage_id:
            key_prefix = (
                SERVICE_NAME
                + ":"
                + DEFAULT_POPULAR_MODULE_NAME.lower()
                + ":"
                + user_label
                + ":"
                + HOMEPAGE_ID_BASED
                + ":"
                + str(homepage_id)
            )
            df_homepage = df_pay_tv[df_pay_tv[HOMEPAGE_ID] == homepage_id]
            df_sum = df_homepage.groupby(CONTENT_ID, as_index=False)[VIEW_COUNT].agg(
                SUM
            )
            df_sum = df_sum.sort_values(by=VIEW_COUNT, ascending=False)
            df_sum.drop(columns=VIEW_COUNT, axis=1, inplace=True)
            df_sum.drop_duplicates(CONTENT_ID, inplace=True)
            additional_df = content_homepage_df[
                content_homepage_df[HOMEPAGE_ID] == homepage_id
            ]
            additional_df = additional_df[
                ~additional_df[CONTENT_ID].isin(unique_ubd_content_id)
            ].reset_index(drop=True)
            additional_df = additional_df.sort_values(by=CREATED_ON, ascending=False)
            additional_df = additional_df[[CONTENT_ID]]
            df_sum = concat([df_sum, additional_df], ignore_index=True)
            df_sum[CREATED_ON] = datetime.utcnow().isoformat()
            df_sum = RecommendationUtils().get_recommendation_scores(df_sum)
            df_sum[REC_TYPE] = DEFAULT_POPULAR_MODULE_NAME.upper()
            if len(df_sum.index) > 0:
                result = df_sum.to_dict(orient=RECORDS)
                homepage_based_rec[key_prefix] = result
        return homepage_based_rec

    @staticmethod
    @custom_exception()
    def get_default_popular_homepage_id():
        PopularModelFallback.clear_data()
        PopularModelFallback.get_data()
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_dict = []
        Logging.info("Get Popular Fallback Content based on Homepage_id")
        start = time.time()
        for user_label in list_user_label:
            Logging.info(
                "Start generating fallback popular content based on Homepage_id for "
                + user_label
                + " users"
            )
            output_dict = PopularModelFallback.default_popular_homepage_id_algorithm(
                user_label
            )
            Logging.info("Success Preparing Fallback Trending Dataframe Output")
            list_output_dict.append(output_dict)
        (
            fallback_popular_pay_tv_homepage_id_dict,
            fallback_popular_no_pay_tv_homepage_id_dict,
        ) = list_output_dict
        end = time.time()
        duration = end - start
        Logging.info(
            "Duration for Get Fallback Popular Content based on Homepage_id is "
            + str(duration)
            + " seconds"
        )
        return (
            fallback_popular_pay_tv_homepage_id_dict,
            fallback_popular_no_pay_tv_homepage_id_dict,
        )

    @staticmethod
    @custom_exception()
    def get_default_popular_all_content():
        PopularModelFallback.get_data()
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_dict = []
        Logging.info("Get Fallback Popular Content based on All Content")
        start = time.time()
        for user_label in list_user_label:
            Logging.info(
                "Start generating fallback popular content based on All Content for "
                + user_label
                + " users"
            )
            output_dict = PopularModelFallback.default_popular_all_content_algorithm(
                user_label
            )
            Logging.info("Success Preparing Fallback Popular Dataframe Output")
            list_output_dict.append(output_dict)
        (
            fallback_popular_pay_tv_all_content_dict,
            fallback_popular_no_pay_tv_all_content_dict,
        ) = list_output_dict
        end = time.time()
        duration = end - start
        Logging.info(
            "Duration for Get Fallback Popular Content based on All Content is "
            + str(duration)
            + " seconds"
        )
        PopularModelFallback.clear_data()
        return (
            fallback_popular_pay_tv_all_content_dict,
            fallback_popular_no_pay_tv_all_content_dict,
        )