import time
from datetime import datetime

import pandas as pd


from offline_results.common.config import (
    CONFIG_HOMEPAGE_PAYTV,
    CONFIG_HOMEPAGE_NO_PAYTV, LIVE_TV_CHANNEL_HOMEPAGE_PAYTV, LIVE_TV_CHANNEL_HOMEPAGE_NO_PAYTV,
)
from offline_results.common.constants import (
    NO_PAY_TV,
    PAY_TV,
    CONTENT_ID,
    SERVICE_NAME,
    ALL_CONTENT_BASED,
    RECORDS,
    DEFAULT_TRENDING_MODULE_NAME,
    HOMEPAGE_ID,
    CREATED_ON,
    VIEW_COUNT,
    SUM,
    CONTENT_STATUS,
    LEFT,
    ACTIVE,
    IS_PAY_TV,
    INDEX,
    MODEL_NAME,
    HOMEPAGE_ID_BASED,
    INNER,
    REC_TYPE,
    TRENDING_MODULE_NAME, SCORE,
)
from offline_results.recommendation.homepage_recommendation.trending.trending_model import (
    TrendingModel,
)

from offline_results.recommendation.utils import RecommendationUtils
from offline_results.repository.graph_db_connection import ANGraphDb
from offline_results.utils import custom_exception, Logging


class TrendingFallback:
    viewed_relation_history_df = pd.DataFrame()
    live_tv_channel_paytv = []
    live_tv_channel_nopaytv = []

    @staticmethod
    @custom_exception()
    def get_data():
        obj = TrendingFallback
        if len(obj.viewed_relation_history_df) != 0:
            return
        obj.viewed_relation_history_df = RecommendationUtils.user_viewed_data_from_s3()
        obj.live_tv_channel_paytv = RecommendationUtils.get_tv_channels(is_pay_tv=True)
        obj.live_tv_channel_nopaytv = RecommendationUtils.get_tv_channels(is_pay_tv=False)

    @staticmethod
    @custom_exception()
    def clear_data():
        obj = TrendingFallback
        obj.viewed_relation_history_df = pd.DataFrame()
        obj.live_tv_channel_paytv = []
        obj.live_tv_channel_nopaytv = []

    @staticmethod
    @custom_exception()
    def default_trending_homepage_algorithm(user_label, graph):

        global df_trending, unique_content_ubd_paytv, unique_content_ubd_nopaytv, df_no_paytv_trend
        df = TrendingFallback.viewed_relation_history_df
        if user_label == PAY_TV:
            df_paytv = df.loc[
                lambda df: (df[IS_PAY_TV] == True) & (df[CONTENT_STATUS] == ACTIVE)
            ]
            try:
                Logging.info("Fetch content-homepage_id mapping")
                (
                    paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=PAY_TV, homepage_id_wise=True
                )
            except Exception:
                graph = ANGraphDb.new_connection_config().graph
                Logging.info("Re-trying to Fetch content-homepage_id mapping")
                (
                    paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=PAY_TV, homepage_id_wise=True
                )
            df_active_homepage = paytv_homepage_df.explode(column=HOMEPAGE_ID)
            unique_content_ubd_paytv = df_paytv[CONTENT_ID].unique().tolist()
            df_paytv_label_merge_with_homepage_id = pd.merge(
                df_paytv, df_active_homepage, on=CONTENT_ID, how=LEFT
            )
            df_paytv_label_merge_with_homepage_id = (
                df_paytv_label_merge_with_homepage_id[
                    ~df_paytv_label_merge_with_homepage_id[HOMEPAGE_ID].isna()
                ].reset_index(drop=True)
            )
            df_trending = pd.DataFrame(CONFIG_HOMEPAGE_PAYTV.items())
            df_trending.columns = [HOMEPAGE_ID, MODEL_NAME]
            df_trending = df_trending[df_trending[MODEL_NAME] == TRENDING_MODULE_NAME.lower()]
            list_dedicated_homepage_id_paytv = df_trending[HOMEPAGE_ID].to_list()
            list_dedicated_homepage_id_paytv.extend(LIVE_TV_CHANNEL_HOMEPAGE_PAYTV)
            df_view_count = df_paytv_label_merge_with_homepage_id[df_paytv_label_merge_with_homepage_id[
                HOMEPAGE_ID].isin(list_dedicated_homepage_id_paytv)
            ].reset_index(drop=True)
        else:
            df_no_paytv = df.loc[
                lambda df: (df[IS_PAY_TV] == False) & (df[CONTENT_STATUS] == ACTIVE)
            ]
            try:
                Logging.info("Fetch content-homepage_id mapping")
                (
                    no_paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=NO_PAY_TV, homepage_id_wise=True
                )
            except Exception:
                graph = ANGraphDb.new_connection_config().graph
                Logging.info("Re-trying to Fetch content-homepage_id mapping")
                (
                    no_paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=NO_PAY_TV, homepage_id_wise=True
                )
            no_paytv_homepage_df = no_paytv_homepage_df.explode(column=HOMEPAGE_ID)
            unique_content_ubd_nopaytv = df_no_paytv[CONTENT_ID].unique().tolist()
            df_no_paytv_label_merge_with_homepage_id = pd.merge(
                df_no_paytv, no_paytv_homepage_df, on=CONTENT_ID, how=LEFT
            )
            df_no_paytv_trending = pd.DataFrame(CONFIG_HOMEPAGE_NO_PAYTV.items())
            df_no_paytv_trending.columns = [HOMEPAGE_ID, MODEL_NAME]
            df_no_paytv_trending = df_no_paytv_trending[df_no_paytv_trending[MODEL_NAME] == TRENDING_MODULE_NAME.lower()]
            list_dedicated_homepage_id_nopaytv = df_no_paytv_trending[HOMEPAGE_ID].to_list()
            list_dedicated_homepage_id_nopaytv.extend(LIVE_TV_CHANNEL_HOMEPAGE_NO_PAYTV)
            df_view_count = df_no_paytv_label_merge_with_homepage_id[df_no_paytv_label_merge_with_homepage_id[
                HOMEPAGE_ID].isin(list_dedicated_homepage_id_nopaytv)].reset_index(drop=True)

        trending_homepage = list_dedicated_homepage_id_paytv if user_label == PAY_TV \
            else list_dedicated_homepage_id_nopaytv
        key_prefix = (
            SERVICE_NAME
            + ":"
            + DEFAULT_TRENDING_MODULE_NAME.lower()
            + ":"
            + user_label
            + ":"
            + HOMEPAGE_ID_BASED
        )
        homepage_based_rec = {}
        for homepage_id in trending_homepage:
            key_prefix_cls = key_prefix + ":" + str(homepage_id)
            filter_df = df_view_count[df_view_count[HOMEPAGE_ID] == homepage_id]
            filter_df = filter_df.groupby(CONTENT_ID, as_index=False)[VIEW_COUNT].agg(
                SUM
            )
            filter_df = filter_df.sort_values(by=VIEW_COUNT, ascending=False)
            additional_df = content_homepage_df[
                content_homepage_df[HOMEPAGE_ID] == homepage_id
            ]
            additional_df = additional_df[~additional_df[CONTENT_ID].isin(
                unique_content_ubd_paytv if user_label == PAY_TV else unique_content_ubd_nopaytv)
            ].reset_index(drop=True)
            additional_df = additional_df.sort_values(by=CREATED_ON, ascending=False)
            additional_df = additional_df[[CONTENT_ID]]
            additional_df[VIEW_COUNT] = 0
            filter_df = pd.concat([filter_df, additional_df], ignore_index=True)
            filter_df[CREATED_ON] = datetime.utcnow().isoformat()
            filter_df[REC_TYPE] = DEFAULT_TRENDING_MODULE_NAME.upper()
            filter_df[VIEW_COUNT] = filter_df[VIEW_COUNT].astype(int)
            filter_df = RecommendationUtils().get_recommendation_scores(filter_df)
            if len(filter_df.index) > 0:
                result = filter_df.to_dict(orient=RECORDS)
                homepage_based_rec[key_prefix_cls] = result
        return homepage_based_rec

    @staticmethod
    @custom_exception()
    def default_trending_all_content_algorithm(user_label, graph):

        df = TrendingFallback.viewed_relation_history_df
        if user_label == PAY_TV:
            df_paytv = df.loc[
                lambda df: (df[IS_PAY_TV] == True) & (df[CONTENT_STATUS] == ACTIVE)
            ]
            try:
                Logging.info("Fetch content-homepage_id mapping")
                (
                    paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=PAY_TV, homepage_id_wise=False
                )
            except Exception:
                graph = ANGraphDb.new_connection_config().graph
                Logging.info("Re-trying to Fetch content-homepage_id mapping")
                (
                    paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=PAY_TV, homepage_id_wise=False
                )
            df2_new = paytv_homepage_df.explode(column=HOMEPAGE_ID)
            df_paytv_label_merge_with_homepage_id = pd.merge(
                df_paytv, df2_new, on=CONTENT_ID, how=LEFT
            )
            df_trending = pd.DataFrame.from_dict(
                CONFIG_HOMEPAGE_PAYTV, orient=INDEX
            ).reset_index()
            df_trending.rename(
                columns={INDEX: HOMEPAGE_ID, 0: MODEL_NAME}, inplace=True
            )
            df_trend = df_trending[df_trending[MODEL_NAME] == TRENDING_MODULE_NAME.lower()]
            df_filter_homepage_id = pd.merge(
                df_paytv_label_merge_with_homepage_id,
                df_trend,
                on=HOMEPAGE_ID,
                how=INNER,
            )
            df_filter_content = df_filter_homepage_id[
                ~df_filter_homepage_id[CONTENT_ID].isin(
                    TrendingFallback.live_tv_channel_paytv
                )
            ]
            df_view_sum = df_filter_content.groupby([CONTENT_ID], as_index=False)[
                VIEW_COUNT
            ].agg(SUM)

            df_homepage_id = df_filter_content[[HOMEPAGE_ID, CONTENT_ID]]
            df_homepage_id = pd.merge(
                df_view_sum, df_homepage_id, on=CONTENT_ID, how=LEFT
            )
        else:
            df_no_paytv = df.loc[
                lambda df: (df[IS_PAY_TV] == False) & (df[CONTENT_STATUS] == ACTIVE)
            ]
            try:
                Logging.info("Fetch content-homepage_id mapping")
                (
                    no_paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=NO_PAY_TV, homepage_id_wise=False
                )
            except Exception:
                graph = ANGraphDb.new_connection_config().graph
                Logging.info("Re-trying to Fetch content-homepage_id mapping")
                (
                    no_paytv_homepage_df,
                    content_homepage_df,
                ) = TrendingModel.get_content_homepage_id_mapping(
                    graph, user_label=NO_PAY_TV, homepage_id_wise=False
                )
            df_no_paytv_homepage = no_paytv_homepage_df.explode(column=HOMEPAGE_ID)
            df_no_paytv_label_merge_with_homepage_id = pd.merge(
                df_no_paytv, df_no_paytv_homepage, on=CONTENT_ID, how=LEFT
            )
            df_no_paytv_trending = pd.DataFrame.from_dict(
                CONFIG_HOMEPAGE_NO_PAYTV, orient=INDEX
            ).reset_index()
            df_no_paytv_trending.rename(
                columns={INDEX: HOMEPAGE_ID, 0: MODEL_NAME}, inplace=True
            )
            df_no_paytv_trend = df_no_paytv_trending[
                df_no_paytv_trending[MODEL_NAME] == TRENDING_MODULE_NAME.lower()
            ]
            df_no_paytv_filter_homepage_id = pd.merge(
                df_no_paytv_label_merge_with_homepage_id,
                df_no_paytv_trend,
                on=HOMEPAGE_ID,
                how=INNER,
            )
            df_no_paytv_filter_content = df_no_paytv_filter_homepage_id[
                ~df_no_paytv_filter_homepage_id[CONTENT_ID].isin(
                    TrendingFallback.live_tv_channel_nopaytv
                )
            ]
            df_view_sum = df_no_paytv_filter_content.groupby(
                [CONTENT_ID], as_index=False
            )[VIEW_COUNT].agg(SUM)

            df_homepage_id = df_no_paytv_filter_content[[HOMEPAGE_ID, CONTENT_ID]]
            df_homepage_id = pd.merge(
                df_view_sum, df_homepage_id, on=CONTENT_ID, how=LEFT
            )
        df_homepage_id = df_homepage_id.drop_duplicates(
            subset=[CONTENT_ID], keep="first"
        )
        filter_df = df_homepage_id.sort_values(by=VIEW_COUNT, ascending=False).iloc[
            :20, :
        ]
        filter_df.drop(columns=VIEW_COUNT, axis=1, inplace=True)
        filter_df.drop_duplicates(CONTENT_ID, inplace=True)
        filter_df[CREATED_ON] = datetime.utcnow().isoformat()
        filter_df[REC_TYPE] = DEFAULT_TRENDING_MODULE_NAME.upper()
        all_content_based_rec = {}
        key_prefix = (
            SERVICE_NAME
            + ":"
            + DEFAULT_TRENDING_MODULE_NAME.lower()
            + ":"
            + user_label
            + ":"
            + ALL_CONTENT_BASED
        )
        result = filter_df.to_dict(orient=RECORDS)
        all_content_based_rec[key_prefix] = result
        return all_content_based_rec

    @staticmethod
    @custom_exception()
    def get_default_trending_homepage_id(graph):
        TrendingFallback.clear_data()
        TrendingFallback.get_data()
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_dict = []
        Logging.info("Get Fallback Trending Content based on Homepage_id")
        start = time.time()
        for user_label in list_user_label:
            Logging.info(
                "Start generating fallback trending content based on Homepage_id for "
                + user_label
                + " users"
            )
            output_dict = TrendingFallback.default_trending_homepage_algorithm(
                user_label, graph
            )
            Logging.info("Success Preparing Fallback Trending Dataframe Output")
            list_output_dict.append(output_dict)
        (
            fallback_trending_pay_tv_homepage_id_dict,
            fallback_trending_no_pay_tv_homepage_id_dict,
        ) = list_output_dict
        end = time.time()
        duration = end - start
        Logging.info(
            "Duration for Get Fallback Trending Content based on Homepage_id is "
            + str(duration)
            + " seconds"
        )
        return (
            fallback_trending_pay_tv_homepage_id_dict,
            fallback_trending_no_pay_tv_homepage_id_dict,
        )

    @staticmethod
    @custom_exception()
    def get_default_trending_all_content(graph):
        TrendingFallback.get_data()
        list_user_label = [PAY_TV, NO_PAY_TV]
        list_output_dict = []
        Logging.info("Get Fallback Trending Content based on All Content")
        start = time.time()
        for user_label in list_user_label:
            Logging.info(
                "Start generating fallback trending content based on All Content for "
                + user_label
                + " users"
            )
            output_dict = TrendingFallback.default_trending_all_content_algorithm(
                user_label, graph
            )
            Logging.info("Success Preparing Fallback Trending Dataframe Output")
            list_output_dict.append(output_dict)
        (
            fallback_trending_pay_tv_all_content_dict,
            fallback_trending_no_pay_tv_all_content_dict,
        ) = list_output_dict
        end = time.time()
        duration = end - start
        Logging.info(
            "Duration for Get Fallback Trending Content based on All Content is "
            + str(duration)
            + " seconds"
        )
        TrendingFallback.clear_data()
        return (
            fallback_trending_pay_tv_all_content_dict,
            fallback_trending_no_pay_tv_all_content_dict,
        )
